#ifndef _PIPEFS_H

#define _PIPEFS_H

#include "fsproto.h"
#include "lists.h"
#include "ringbuff.h"
#include "lock.h"

#include <Errors.h>
#include <sys/stat.h>
#include <SupportDefs.h>

/* this flag passed makes the pipe permanent (i.e. not deleted on the last close) */
#define O_MKFIFO 0x10000

/* should be more than enough */
#define PIPEFS_NAME_LEN B_OS_NAME_LENGTH
/* this is quite a standard value */
#define PIPEFS_BUFF_SIZE 4096

#define PIPEFS_NAME "npipefs"

#define ASSERT(op) if (!(op)) panic("ASSERT: %s in %s:%s", #op, __FILE__, __FUNCTION__)

struct mount_fs_params
{
	/* no param */
};

struct fs_file_cookie;

struct fs_node
{
	slist_entry lent; /* must stay here */
	char name[PIPEFS_NAME_LEN];
	vnode_id vnid;
	lock l;
	bool is_perm;	/* don't delete on last close */
	bool got_wr;	/* set when at least 1 writer has been there */
	struct stat st; /* including S_IFDIR in st_mode */
	struct fs_file_cookie *readers;
	struct fs_file_cookie *writers;
	struct _slist_entry *waiting_readers;
	struct _slist_entry *waiting_writers;
	struct ring_buffer buffer; /* must stay last */
};

struct fs_nspace
{
	nspace_id nsid;
	vnode_id rootid;
	vnode_id nextid;
	
	lock l;
	
	struct fs_node *root; /* fast access for stat time change */
	struct fs_node *first; /* linked list */
};

#define FS_WAIT_NONE 0
#define FS_WAIT_SEM 1
#define FS_WAIT_SEL 2

#define SELECT_MAX SELECT_EXCEPTION+1

struct fs_file_cookie
{
	slist_entry lent; /* must stay here */
	slist_entry wlent; /* wait list entry */
	struct fs_node *node;
	int dir_current; /* current entry for readdir() */
	int omode;
	int wflags; /* waiting on what */
	sem_id sem;
	selectsync *selsync[SELECT_MAX]; /* indexed by SELECT_* */
	uint32 selref[SELECT_MAX];
};

typedef struct fs_nspace fs_nspace;
typedef struct fs_node fs_node;
typedef struct fs_file_cookie fs_file_cookie;
/* not much different */
#define fs_dir_cookie fs_file_cookie

vnode_id new_vnid(fs_nspace *ns);

#define FS_NOTIFY_READER 0x0001
#define FS_NOTIFY_WRITER 0x0002
#define FS_NOTIFY_ALL 0x0010
int pipefs_event(fs_nspace *ns, fs_node *node, int flags);

int	pipefs_read_vnode(fs_nspace *ns, vnode_id vnid, char r, fs_node **node);
int	pipefs_write_vnode(fs_nspace *ns, fs_node *node, char r);
int	pipefs_remove_vnode(fs_nspace *ns, fs_node *node, char r);
int	pipefs_secure_vnode(fs_nspace *ns, fs_node *node);

int	pipefs_walk(fs_nspace *ns, fs_node *base, const char *file, char **newpath,
					vnode_id *vnid);

int pipefs_opendir(fs_nspace *ns, fs_node *node, fs_dir_cookie **cookie);
int	pipefs_closedir(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie);
int	pipefs_rewinddir(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie);
int	pipefs_readdir(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie, long *num,
					struct dirent *buf, size_t bufsize);

int pipefs_free_dircookie(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie);
int	pipefs_rstat(fs_nspace *ns, fs_node *node, struct stat *);
int	pipefs_mount(nspace_id nsid, const char *devname, ulong flags,
					struct mount_fs_params *parms, size_t len, fs_nspace **data, vnode_id *vnid);
int	pipefs_unmount(fs_nspace *ns);
int pipefs_rfsstat(fs_nspace *ns, struct fs_info *info);

int	pipefs_open(fs_nspace *ns, fs_node *node, int omode, fs_file_cookie **cookie);
int	pipefs_close(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie);
int pipefs_free_cookie(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie);
int pipefs_read(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, off_t pos, void *buf,
					size_t *len);

int pipefs_write(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, off_t pos,
					const void *buf, size_t *len);

int pipefs_wstat(fs_nspace *ns, fs_node *node, struct stat *st, long mask);
int pipefs_wfsstat(fs_nspace *ns, struct fs_info *info, long mask);
int	pipefs_create(fs_nspace *ns, fs_node *dir, const char *name,
					int omode, int perms, vnode_id *vnid, fs_file_cookie **cookie);

int	pipefs_unlink(fs_nspace *ns, fs_node *dir, const char *name);

int	pipefs_mkdir(fs_nspace *ns, fs_node *dir, const char *name,	int perms);

int	pipefs_rename(fs_nspace *ns, fs_node *olddir, const char *oldname,
					fs_node *newdir, const char *newname);

int	pipefs_rmdir(fs_nspace *ns, fs_node *dir, const char *name);
int	pipefs_readlink(fs_nspace *ns, fs_node *node, char *buf, size_t *bufsize);

int	pipefs_symlink(fs_nspace *ns, fs_node *dir, const char *name,
					const char *path);

extern int pipefs_access(fs_nspace *ns, fs_node *node, int mode);
extern int pipefs_setflags(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, int flags);

extern int pipefs_select(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, uint8 event, uint32 ref, selectsync *sync);
extern int pipefs_deselect(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, uint8 event, selectsync *sync);

#endif
